// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

part of ducene;

abstract class IndexDirectory {
  void initOutput();
  void outputIndex(String name, Map<String,DocSet> index);
  void outputData(String name, Map<int,DocData> data);
  void initInput();
  IndexSource inputIndex();
  DataSource inputData();
}
abstract class IndexStreams {
  void readFully(Map<String,Map<String,DocSet>> to);
  void readTerms(String field, Map<String,Map<String,DocSet>> to);
  void readPostings(String field, String term, Map<String,Map<String,DocSet>> to);
}
abstract class DataStreams {
  void readFully(Map<String,Map<int,DocData>> to);
  void readFields(Map<String,Map<int,DocData>> to);
  void readValues(String field, int docId, Map<String,Map<int,DocData>> to);
}
class DocSet {
  Uint32List buffer;
  DocSet(Set<int> docIds) {
    buffer = new Uint32List.fromList(docIds.toList());
  }
  Set<int> toSet() {
    return buffer.toSet();
  }
}
class DocData {
  Uint16List buffer;
  DocData(List<String> data) {
    buffer = new Uint16List.fromList(data.join('\$').codeUnits);
  }
  List<String> toList() {
    return new String.fromCharCodes(buffer).split('\$');
  }
}

class RAMIndexDirectory implements IndexDirectory {

  Map<String,Map<String,DocSet>> _ramIndex = {};
  Map<String,Map<int,DocData>> _ramData = {};

  void initOutput() { }

  void outputIndex(String name, Map<String,DocSet> index) {
    _ramIndex[name] = index;
  }

  void outputData(String name, Map<int,DocData> data) {
    _ramData[name] = data;
  }

  void initInput() { }
  IndexSource inputIndex() => new IndexSource(new RAMIndexStreams(_ramIndex));
  DataSource inputData() => new DataSource(new RAMDataStreams(_ramData));

}
class RAMIndexStreams extends IndexStreams {
  final Map<String,Map<String,DocSet>> _ramIndex;
  RAMIndexStreams(this._ramIndex);
  bool _readFully = false;
  void readFully(Map to) {
    for (String field in _ramIndex.keys) {
      to[field] = _ramIndex[field];
    }
    _readFully = true;
  }
  void readTerms(String field, Map to) {
    if (!_readFully) {
      readFully(to);
    }
  }
  void readPostings(String field, String term, Map to) {
    if (!_readFully) readFully(to);
  }
}
class RAMDataStreams extends DataStreams {
  final Map<String,Map<int,DocData>> _ramData;
  RAMDataStreams(this._ramData);
  bool _readFully = false;
  void readFully(Map to) {
    for (String field in _ramData.keys) {
      to[field] = _ramData[field];
    }
    _readFully = true;
  }
  void readFields(Map to) {
    if (!_readFully) readFully(to);
  }
  void readValues(String field, int docId, Map to) {
    if (!_readFully) readFully(to);
  }
}

class FSIndexDirectory implements IndexDirectory {

  final Directory _dir;

  FSIndexDirectory(this._dir);

  void initOutput() {
    if (_dir.existsSync()) _dir.deleteSync(recursive: true);
    if (!_dir.existsSync()) _dir.createSync(recursive: true);
  }

  void outputIndex(String name, Map<String,DocSet> index) {
    File indexFile = new File('${_dir.path}/${name}.idx');
    if (!indexFile.existsSync()) indexFile.createSync();
    for (String key in index.keys) {
      String value = index[key].toSet().join(',');
      String line = '$key=>$value';
      indexFile.writeAsStringSync(line+'\n', mode: FileMode.APPEND, encoding: UTF8);
    }
  }

  void outputData(String name, Map<int,DocData> data) {
    File storeFile = new File('${_dir.path}/${name}.data');
    if (!storeFile.existsSync()) storeFile.createSync();
    for (int key in data.keys) {
      String value = data[key].toList().join('\$');
      String line = '${key.toString()}=>${value}';
      storeFile.writeAsStringSync(line+'\n', mode: FileMode.APPEND, encoding: UTF8);
    }
  }

  void initInput() { }

  IndexSource inputIndex() => new IndexSource(new FSIndexStreams(_dir));
  DataSource inputData() => new DataSource(new FSDataStreams(_dir));

}
class FSIndexStreams extends IndexStreams {
  final Directory _dir;
  FSIndexStreams(this._dir);
  bool _readFully = false;
  void readFully(Map to) {
    for (String path in FSIndexUtils.getFileList(_dir, '.idx')) {
      String name = FSIndexUtils.getName(path, '.idx');
      File indexFile = new File(path);
      List<String> lines = indexFile.readAsLinesSync();
      var invertedIndex = <String,DocSet>{};
      for (String line in lines) {
        List<String> kv = line.split('=>');
        String key = kv[0];
        Set<int> values = kv[1].split(',').map((s) => int.parse(s)).toSet();
        invertedIndex[key] = new DocSet(values);
      }
      to[name] = invertedIndex;
    }
    _readFully = true;
  }
  void readTerms(String field, Map to) {
    if (!_readFully) readFully(to);
  }
  void readPostings(String field, String term, Map to) {
    if (!_readFully) readFully(to);
  }
}
class FSDataStreams extends DataStreams {
  final Directory _dir;
  FSDataStreams(this._dir);
  bool _readFully = false;
  void readFully(Map to) {
    for (String path in FSIndexUtils.getFileList(_dir, '.data')) {
      String name = FSIndexUtils.getName(path, '.data');
      File dataFile = new File(path);
      String str = dataFile.readAsStringSync();
      List<String> lines = str.split('\n');
      var storedData = <int,DocData>{};
      for (String line in lines) {
        List<String> kv = line.split('=>');
        if (kv.length == 2) {
          int key = int.parse(kv[0]);
          List<String> values = kv[1].split('\$');
          storedData[key] = new DocData(values);
        }
      }
      to[name] = storedData;
    }
    _readFully = true;
  }
  void readFields(Map to) {
    if (!_readFully) readFully(to);
  }
  void readValues(String field, int docId, Map to) {
    if (!_readFully) readFully(to);
  }
}
class FSIndexUtils {
  static Iterable<String> getFileList(Directory dir, String extension) =>
      dir.listSync(recursive: false, followLinks: false)
          .map((e) => e.path)
          .where((path) => path.endsWith(extension));

  static String getName(String path, String extension) =>
      path.split('/').last.replaceAll(extension, '');
}

class FSIndexJsonConverter {
  final Directory _dir;
  FSIndexDirectory _fsDir;

  FSIndexJsonConverter(this._dir) {
    _fsDir = new FSIndexDirectory(_dir);
  }

  void run() {
    Map ret = {};

    Map<String,Map<String,DocSet>> inputIndex = {};
    inputIndex = _fsDir.inputIndex().read(inputIndex);
    Map<String,Map<String,Map<String,String>>> idx = {};
    for (String fieldName in inputIndex.keys) {
      Map<String,DocSet> fieldValue = inputIndex[fieldName];
      Map<String,Map<String,String>> terms = {};
      int i = 0;
      for (String term in fieldValue.keys) {
        Map<String,String> termInfo = {};
        Set<int> postings = fieldValue[term].toSet();
        termInfo["postings"] = postings.join(",");
        termInfo["term"] = term;
        terms[i.toString()] = termInfo;
        i++;
      }
      idx[fieldName] = terms;
    }

    Map<String,Map<int,DocData>> inputData = {};
    inputData = _fsDir.inputData().read(inputData);
    Map<String,Map<String,String>> data = {};
    for (String fieldName in inputData.keys) {
      Map<int,DocData> fieldValue = inputData[fieldName];
      Map<String,String> fieldData = {};
      for (int docId in fieldValue.keys) {
        List<String> data = fieldValue[docId].toList();
        fieldData[docId.toString()] = data.join("\$");
      }
      data[fieldName] = fieldData;
    }

    ret["idx"] = idx;
    ret["data"] = data;

    _outJson(ret, "fsindex.json");
  }

  void _outJson(Map obj, String outFilePath) {
    File outFile = new File(outFilePath);
    if (outFile.existsSync()) outFile.deleteSync();
    if (!outFile.existsSync()) outFile.createSync();
    outFile.writeAsStringSync(JSON.encode(obj), mode: FileMode.APPEND, encoding: UTF8);
  }

}