// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

part of ducene;

class IndexReader {

  IndexSource _indexSource;
  DataSource _dataSource;
  Map<String,Map<String,DocSet>> _indexes = <String,Map<String,DocSet>>{};
  Map<String,Map<int,DocData>> _data = <String,Map<int,DocData>>{};

  IndexReader(IndexDirectory directory) {
    _indexSource = directory.inputIndex();
    _dataSource = directory.inputData();
  }

  void read() {
    _readIndexes();
    _readData();
  }

  Map<String,Map<String,DocSet>> _readIndexes() {
    if (_indexes.length == 0) {
      _indexes = _indexSource.read(_indexes);
    }
    return _indexes;
  }

  Iterable<String> terms(String field) {
    if (_indexes.length == 0 || _indexes[field] == null) {
      _indexes = _indexSource.terms(field, _indexes);
    }
    return (_indexes[field] == null) ? [] : _indexes[field].keys;
  }

  DocSet postings(String field, String term) {
    if (terms(field).length == 0) return new DocSet(new Set());
    _indexes = _indexSource.postings(field, term, _indexes);
    DocSet r = _indexes[field][term];
    return (r == null) ? new DocSet(new Set()) : r;
  }

  Map<String,Map<int,DocData>> _readData() {
    if (_data.length == 0) {
      _data = _dataSource.read(_data);
    }
    return _data;
  }

  Iterable<String> dataFields() {
    if (_data.length == 0) {
      _data = _dataSource.fields(_data);
    }
    return _data.keys;
  }

  List<String> dataValues(String field, int docId) {
    if (_data.length == 0 || _data[field] == null || _data[field][docId] == null) {
      _data = _dataSource.values(field, docId, _data);
    }
    return (_data[field] == null || _data[field][docId] == null)
        ? [] : _data[field][docId].toList();
  }

  int docFreq(String field, String text) {
    Set<int> r = postings(field, text).toSet();
    return (r == null) ? -1 : r.length;
  }
  int maxDoc() {
    List<String> r = terms('id').toList();
    return (r == null) ? 0 : r.length;
  }

  Document document(int docId, {Iterable<String> fieldsToLoad}) {
    var fields = (fieldsToLoad != null) ? fieldsToLoad : dataFields();
    Document ret = new Document();
    for (String field in fields) {
      var text = dataValues(field, docId);
      ret.add(new Field(field, text, null, null));
    }
    return ret;
  }

  @deprecated
  Map<String,Set<int>> getTerms(String field) => throw new Exception('deprecated');
  @deprecated
  Map<int,List<String>> getData(String field) => throw new Exception('deprecated');

}
class IndexSource {
  final IndexStreams _streams;
  IndexSource(this._streams);
  Map<String,Map<String,DocSet>> read(Map<String,Map<String,DocSet>> to) {
    _streams.readFully(to);
    return to;
  }
  Map<String,Map<String,DocSet>> terms(String field, Map<String,Map<String,DocSet>> to) {
    _streams.readTerms(field, to);
    return to;
  }
  Map<String,Map<String,DocSet>> postings(String field, String term, Map<String,Map<String,DocSet>> to) {
    _streams.readPostings(field, term, to);
    return to;
  }
}
class DataSource {
  DataStreams _streams;
  DataSource(this._streams);
  Map<String,Map<int,DocData>> read(Map<String,Map<int,DocData>> to) {
    _streams.readFully(to);
    return to;
  }
  Map<String,Map<int,DocData>> fields(Map<String,Map<int,DocData>> to) {
    _streams.readFields(to);
    return to;
  }
  Map<String,Map<int,DocData>> values(String field, int docId, Map<String,Map<int,DocData>> to) {
    _streams.readValues(field, docId, to);
    return to;
  }
}
class DirectoryReader {
  static IndexReader open(IndexDirectory directory) => new IndexReader(directory)..read();
}