// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

part of ducene;

class BoolQuery {
  final Op _op;
  BoolQuery(this._op);
  List<TermQuery> _termQueries = [];
  List<Query> _filters = [];
  BoolQuery addQuery(TermQuery tQuery) {
    _termQueries.add(tQuery);
    return this;
  }
  BoolQuery addFilter(Query query) {
    _filters.add(query);
    return this;
  }
  static Analyzer defaultAnalyzer = new KeywordAnalyzer();
  BoolQuery append(String field, String text,
      {Analyzer analyzer, double boost: 1.00}) {
    Analyzer ana = (analyzer == null)? defaultAnalyzer : analyzer;
    for (var token in ana.getTokens(text)) {
      addQuery(new TermQuery(new Term(field, token, boost: boost)));
    }
    return this;
  }
  List<TermQuery> get termQueries => _termQueries;
  List<Query> get filters => _filters;
  Op get op => _op;
}
enum Op {
  and, or
}
abstract class Query {
  Set<int> _matchedDocIds;
  Set<int> get matchedDocIds => _matchedDocIds;
  Set<int> match(IndexSearcher searcher) {
    _matchedDocIds = _matchDocIds(searcher);
    return _matchedDocIds;
  }
  Set<int> _matchDocIds(IndexSearcher searcher);
  double score();
}
class TermQuery extends Query {
  final Term _term;
  TermQuery(this._term);
  Term get term => _term;
  Set<int> _matchDocIds(IndexSearcher searcher) {
    return searcher.reader.postings(_term.field, _term.text).toSet();
  }
  double score() => (1.00 * _term.boost);
}
class Term {
  final String field;
  final String text;
  double boost;
  Term(this.field, this.text, {double boost: 1.00}) {
    this.boost = boost;
  }
}
class RangeQuery extends Query {
  final String field;
  final double lower;
  final double upper;
  RangeQuery(this.field, this.lower, this.upper);
  Set<int> _matchDocIds(IndexSearcher searcher) {
    Set<int> ret = new Set();
    Iterable<String> terms = searcher.reader.terms(field);
    if (terms == null) return ret;
    bool isFirst = true;
    for (String term in terms) {
      double value = double.parse(term);
      Set<int> docIds = new Set();
      if (lower <= value && value <= upper) {
        docIds = searcher.reader.postings(field, term).toSet();
      }
      if (isFirst) {
        ret = docIds;
        isFirst = false;
        continue;
      }
      ret = IndexSearcher.unionDocSet(Op.or, ret, docIds);
    }
    return ret;
  }
  double score() => 0.0;
}
class TermsQuery extends Query {
  final List<Term> _terms;
  TermsQuery(this._terms);
  Set<int> _matchDocIds(IndexSearcher searcher) {
    Set<int> orIds = new Set();
    for (Term term in _terms) {
      TermQuery tq = new TermQuery(term);
      Set<int> docIds = tq.match(searcher);
      orIds = IndexSearcher.unionDocSet(Op.or, orIds, docIds);
    }
    return orIds;
  }
  double score() => 0.00;
}
class NotQuery extends Query {
  final Term _allTerm;
  final List<Term> _prohibitTerms;
  NotQuery(this._allTerm, this._prohibitTerms);
  Set<int> _matchDocIds(IndexSearcher searcher) {
    Set<int> prohibitedIds = new Set();
    TermsQuery tsq = new TermsQuery(_prohibitTerms);
    prohibitedIds = tsq.match(searcher);
    TermQuery tq = new TermQuery(_allTerm);
    Set<int> allIds = tq.match(searcher);
    return allIds.difference(prohibitedIds);
  }
  double score() => 0.00;
}